#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re

from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, Method, XCFunctionals, System,\
    SingleConfigurationCalculation, SamplingMethod, ScfIteration

from nwchemparser.metainfo import m_env
from nwchemparser.metainfo.nwchem import x_nwchem_section_start_information,\
    x_nwchem_section_qmd_step


def fix_dfloat(val_in):
    return float(re.sub(r'[dD]([\-\+]*\d)', r'E\1', val_in))


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.\-\+ED]+'

        def str_to_functional(val_in):
            val = [v.strip() for v in val_in.strip().rsplit('  ', 1)]
            if len(val) == 2:
                val = [val[0]] + val[1].split()

            return val

        def str_to_energy(val_in):
            separator = '=' if '=' in val_in else ':'
            val = [v.strip() for v in val_in.strip().split(separator)]
            return val[0], float(val[1]) * ureg.hartree

        def str_to_labels_positions(val_in):
            val = [v.strip().split() for v in val_in.strip().split('\n')]
            labels, positions = [], []
            for val_i in val:
                labels.append(val_i[1])
                positions.append(val_i[3: 6])
            positions = np.array(positions, dtype=np.dtype(np.float64))
            return labels, positions * ureg.angstrom

        def str_to_labels_positions_forces(val_in):
            val = np.transpose([v.split() for v in val_in.strip().split('\n')])
            labels = val[0]
            positions = np.transpose(np.array(val[2:5], dtype=np.dtype(np.float64)))
            forces = np.transpose(np.array(val[5:8], dtype=np.dtype(np.float64)))
            return labels, positions * ureg.bohr, forces * ureg.hartree / ureg.bohr

        geometry_quantities = [
            Quantity(
                'labels_positions',
                r'No\.\s*Tag\s*Charge\s*X\s*Y\s*Z[\s\-]+([\s\S]+?)\n *\n',
                str_operation=str_to_labels_positions, convert=False),
            Quantity(
                'lattice_vectors',
                r'a1\=\<([\d\.\- ]+)\>\s*'
                r'a2\=\<([\d\.\- ]+)\>\s*'
                r'a3\=\<([\d\.\- ]+)\>\s*',
                dtype=np.dtype(np.float64), shape=(3, 3), unit=ureg.bohr)
        ]

        dft_quantities = geometry_quantities + [
            Quantity(
                'general_info',
                r'General Information\s+\-+([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'info',
                    r' +([\w\. ]+\s*\:\s*[\w\.\- ]+)',
                    str_operation=lambda x: [v.strip() for v in x.split(':')],
                    repeats=True)])),
            Quantity(
                'xc_info',
                r'XC Information\s+\-+([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'functional',
                    r'(.+(?:Functional|Exchange|POtential).+)',
                    str_operation=str_to_functional, repeats=True)])),
            Quantity(
                'energy',
                rf'([\w \-\.]+ energy\s*=\s*{re_float})',
                str_operation=str_to_energy, repeats=True, convert=False),
            Quantity(
                'labels_positions_forces',
                r'atom\s*coordinates\s*gradient\s*x\s*y\s*z\s*x\s*y\s*z\s*([\s\S]+?\n *\n)',
                str_operation=str_to_labels_positions_forces, convert=False),
            Quantity(
                'qmd_info',
                r'QMD Run Information\s*\-+\s*([\s\S]+?)\n *\n',
                str_operation=lambda x: [[vi.strip() for vi in v.split(':')] for v in x.split('\n')],
                convert=False),
            Quantity(
                'self_consistency',
                r'convergence\s*iter\s*energy\s*DeltaE\s*RMS\-Dens\s*Diis\-err\s*time\s*'
                r'([\s\S]+?)\n *\n *\n',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        rf'd\=\s*\d+,ls=[\d\.]+,diis\s*\d+\s*({re_float})\s*({re_float})\s*{re_float}\s*{re_float}\s*({re_float})',
                        dtype=np.dtype(np.float64), repeats=True)])),
        ]

        dft_gradient_quantities = [
            Quantity(
                'labels_positions_forces',
                r'atom\s*coordinates\s*gradient\s*x\s*y\s*z\s*x\s*y\s*z\s*([\s\S]+?\n *\n)',
                str_operation=str_to_labels_positions_forces, convert=False),
            Quantity(
                'energy',
                r'\@ Step\s*Energy.+\s*\@[\-\s]+\@\s*(.+)'),
        ]

        pw_quantities = geometry_quantities + [
            Quantity(
                'energy',
                rf'([\w \-]+ energy\s*\:\s*{re_float})',
                str_operation=str_to_energy, repeats=True, convert=False),
            Quantity(
                'spin_S2',
                rf'\<S\^2\>\s*\=\s*({re_float})', dtype=np.float64),
            Quantity(
                'parameters',
                r'options\:\s*([\s\S]+?)\n *\n',
                str_operation=lambda x: [[vi.strip() for vi in v.split('=')] for v in x.split('\n')]),
            Quantity(
                'self_consistency',
                r'ITERATION STARTED([\s\S]+?)ITERATION ENDED',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        rf'\d+\s+({re_float}\s+{re_float})\s+{re_float}\n',
                        repeats=True, dtype=np.dtype(np.float64))])),
            Quantity('total_charge', rf'total charge\:\s*({re_float})', dtype=np.float64)
            # TODO add xc functionals cannot find mapping
        ]

        calculation_quantities = geometry_quantities + [
            Quantity(
                'dft',
                r'(DFT Module[\s\S]+?)(?:NWChem|\Z)',
                repeats=True, sub_parser=TextParser(quantities=dft_quantities)),
            Quantity(
                'dft_gradient',
                r'(DFT Gradient Module[\s\S]+?)(?:NWChem|\Z)',
                repeats=True, sub_parser=TextParser(quantities=dft_gradient_quantities)),
            Quantity(
                'pw',
                r'(PSPW Calculation[\s\S]+?)(?:\*\s*NWPW|\Z)',
                repeats=True, sub_parser=TextParser(quantities=pw_quantities))]

        self._quantities = [
            Quantity(
                'version',
                r'Northwest Computational Chemistry Package \(NWChem\) (.+)\n',
                flatten=False, convert=False),
            Quantity(
                'job_info',
                r'Job information\s*\-+\s*([\s\S]+?)\n *\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'info',
                    r'(\w.+?\s*\=\s*.+)',
                    str_operation=lambda x: [v.strip() for v in x.split('=')],
                    convert=False, repeats=True)])),
            Quantity(
                'input',
                r'(Input Module[\s\S]+? {30}NWChem)',
                sub_parser=TextParser(quantities=geometry_quantities)),
            Quantity(
                'single_point',
                r'(DFT Module\s+\-+[\s\S]+?)(?:GA Statistics for process|NWChem Input Module|\Z)',
                sub_parser=TextParser(quantities=calculation_quantities)),
            Quantity(
                'geometry_optimization',
                r'(Geometry Optimization[\s\S]+?)(?:GA Statistics for process|NWChem Input Module|\Z)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'parameters',
                        r'(maximum gradient threshold\s*\(gmax\)[\s\S]+?)\n *\n',
                        str_operation=lambda x: [[vi.strip() for vi in v.split('=')] for v in x.split('\n')]),
                    Quantity(
                        'iteration',
                        r'p\s+\d+\s*\-+\s*(Geometry[\s\S]+?(?:\-+\s*Ste|\Z))',
                        repeats=True, sub_parser=TextParser(quantities=calculation_quantities))])),
            Quantity(
                'molecular_dynamics',
                r'(QMD Module[\s\S]+?)(?:GA Statistics for process|NWChem Input Module|\Z)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'parameters',
                        r'QMD Run Parameters\s*\-+\s*([\s\S]+?)\n *\n',
                        str_operation=lambda x: [[vi.strip() for vi in v.split(':')] for v in x.split('\n')]),
                    Quantity(
                        'iteration',
                        r'(DFT Module[\s\S]+?(?:NWChem|\Z))',
                        repeats=True, sub_parser=TextParser(quantities=calculation_quantities))])),
            Quantity(
                'pw',
                r'(PSPW Calculation[\s\S]+?)(?:\*\s*NWPW|GA Statistics for process|\Z)',
                repeats=True, sub_parser=TextParser(quantities=pw_quantities))
            # TODO implement frequency analysis
            # TODO add timings
        ]


class NWChemParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/nwchem', code_name='NWChem', code_homepage='http://www.nwchem-sw.org/',
            mainfile_contents_re=(
                r'Northwest Computational Chemistry Package \(NWChem\) (\d+\.)+\d+'))

        self._metainfo_env = m_env

        self.out_parser = OutParser()

        self._metainfo_map = {
            'hostname': 'run_host_name', 'program': 'program_name', 'date': 'start_datetime',
            'compiled': 'compilation_datetime', 'nwchem branch': 'branch',
            'nwchem revision': 'revision', 'input': 'input_filename', 'prefix': 'input_prefix',
            'data base': 'db_filename', 'Charge': 'total_charge',
            'Spin multiplicity': 'spin_target_multiplicity',
            'Maximum number of iterations': 'scf_max_iteration',
            'Convergence on energy requested': 'scf_threshold_energy_change',
            'Total DFT energy': 'energy_total', 'One electron energy': 'x_nwchem_energy_one_electron',
            'Coulomb energy': 'x_nwchem_energy_coulomb', 'Exchange energy': 'energy_X',
            'Exchange-Corr. energy': 'energy_XC',
            'Correlation energy': 'energy_C', 'Nuclear repulsion energy': 'x_nwchem_energy_nuclear_repulsion',
            'total     energy': 'energy_total', 'exc-corr  energy': 'energy_XC',
            # TODO verify if energy contributions mapping is correct
            # 'total orbital energy': 'energy_sum_eigenvalues'
            'hartree   energy': 'energy_correction_hartree',
            'ion-ion   energy': 'x_nwchem_energy_nuclear_repulsion',
            'No. of nuclear steps': 'qmd_number_of_nuclear_steps',
            'Nuclear time step': 'qmd_nuclear_time_step',
            'Target temp. (K)': 'qmd_target_temperature', 'Thermostat': 'qmd_thermostat',
            'Tau': 'qmd_tau', 'Random seed': 'qmd_random_seed',
            'Nuclear integrator': 'qmd_nuclear_integrator', 'Current temp. (K)': 'qmd_initial_temperature',
            'electron spin': 'electron_spin_restriction'}

        self._xc_functional_map = {
            'B3LYP Method XC Potential': 'HYB_GGA_XC_B3LYP',
            'PBE0 Method XC Functional': 'HYB_GGA_XC_PBEH',
            'Becke half-and-half Method XC Potential': 'HYB_GGA_XC_BHANDH',
            'HCTH120  Method XC Functional': 'GGA_XC_HCTH_120',
            'HCTH147  Method XC Functional': 'GGA_XC_HCTH_147',
            'HCTH407 Method XC Functional': 'GGA_XC_HCTH_407',
            'PerdewBurkeErnzerhof Exchange Functional': 'GGA_X_PBE',
            'Becke 1988 Exchange Functional': 'GGA_X_B88',
            'Lee-Yang-Parr Correlation Functional': 'GGA_C_LYP',
            'Perdew 1986 Correlation Functional': 'GGA_C_P86',
            'Perdew 1991 Correlation Functional': 'GGA_C_PW91',
            'Perdew 1991   Exchange Functional': 'GGA_X_PW91',
            'Hartree-Fock (Exact) Exchange': 'HF_X',
            'TPSS metaGGA Exchange Functional': 'MGGA_X_TPSS',
            'TPSS03 metaGGA Correlation Functional': 'MGGA_C_TPSS',
            'Slater Exchange Functional': 'LDA_X',
            'VWN V Correlation Functional': 'LDA_C_VWN',
            'OPTX     Exchange Functional_1.432_non-local+Slater Exchange Functional_1.052_local': 'GGA_X_OPTX',
            'Perdew 1991 LDA Correlation Functional_1.000_local+PerdewBurkeErnz. Correlation Functional_1.000_non-local': 'GGA_C_PBE',
            'Perdew 1981 Correlation Functional_1.000_local+Perdew 1986 Correlation Functional_1.000_non-local': 'GGA_C_P86',
            'Perdew 1991 Correlation Functional_1.000_non-local+Perdew 1991 LDA Correlation Functional_1.000_local': 'GGA_C_PW91'
        }

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger

    def parse_system(self, source):
        sec_system = self.archive.section_run[0].m_create(System)

        labels, positions = source.get('labels_positions', [[], []])
        if len(positions) == 0:
            dft = source.get('dft', [{}])[-1]
            dft_gradient = source.get('dft_gradient', [{}])[-1]
            labels_positions = dft.get('labels_positions_forces', dft_gradient.get('labels_positions_forces'))
            if labels_positions is not None:
                labels, positions = labels_positions[:2]

        if len(positions) == 0:
            labels, positions = self.out_parser.get('input', {}).get('labels_positions', [[], []])

        sec_system.atom_labels = labels
        sec_system.atom_positions = positions

        lattice_vectors = source.get('lattice_vectors')
        if lattice_vectors is not None:
            sec_system.lattice_vectors = lattice_vectors
            sec_system.configuration_periodic_dimensions = [True, True, True]
        else:
            sec_system.configuration_periodic_dimensions = [False, False, False]

        return sec_system

    def parse_scc(self, source):
        sec_scc = self.archive.section_run[0].m_create(SingleConfigurationCalculation)

        # we only read the results from the last dft and gradients module
        dft = source.get('dft', [{}])[-1]
        dft_gradient = source.get('dft_gradient', [{}])[-1]

        # energies
        for key, val in source.get('energy', dft.get('energy', [])):
            key = self._metainfo_map.get(key)
            if key is not None:
                setattr(sec_scc, key, val.to('J').magnitude)

        # for geometry optimization, energy can also be parsed from gradient module
        energy = dft_gradient.get('energy')
        if energy is not None:
            sec_scc.energy_total = energy[1] * ureg.hartree
            sec_scc.time_calculation = energy[-1]

        # forces
        forces = dft_gradient.get('labels_positions_forces', dft.get('labels_positions_forces'))
        if forces is not None:
            sec_scc.atom_forces = forces[2]

        # spin
        spin_S2 = source.get('spin_S2')
        if spin_S2 is not None:
            sec_scc.spin_S2 = spin_S2

        # md properties
        qmd_info = dft.get('qmd_info')
        if qmd_info is not None:
            sec_qmd = sec_scc.m_create(x_nwchem_section_qmd_step)
            for key, val in qmd_info:
                val = val.split()
                if key == 'Time elapsed (fs)':
                    sec_qmd.x_nwchem_qmd_step_time = float(val[0]) * ureg.fs
                elif key == 'Kin. energy (a.u.)':
                    sec_qmd.x_nwchem_qmd_step_kinetic_energy = float(val[1]) * ureg.hartree
                elif key == 'Pot. energy (a.u.)':
                    sec_qmd.x_nwchem_qmd_step_potential_energy = float(val[1]) * ureg.hartree
                elif key == 'Tot. energy (a.u.)':
                    sec_qmd.x_nwchem_qmd_step_total_energy = float(val[1]) * ureg.hartree
                elif key == 'Target temp. (K)':
                    sec_qmd.x_nwchem_qmd_step_target_temperature = float(val[1]) * ureg.kelvin
                elif key == 'Current temp. (K)':
                    sec_qmd.x_nwchem_qmd_step_temperature = float(val[1]) * ureg.kelvin
                elif key == 'Dipole (a.u.)':
                    sec_qmd.x_nwchem_qmd_step_dipole = [float(v) for v in val[1:4]]

        # self consistency
        scf = dft.get('self_consistency', source.get('self_consistency'))
        if scf is not None:
            for iteration in scf.get('iteration', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                iteration = [fix_dfloat(i) if isinstance(i, str) else i for i in iteration]
                sec_scf.energy_total_scf_iteration = iteration[0] * ureg.hartree
                sec_scf.energy_change_scf_iteration = iteration[1] * ureg.hartree
                if len(iteration) > 2:
                    sec_scf.time_scf_iteration = iteration[2]

        return sec_scc

    def parse_method(self, source):
        sec_method = self.archive.section_run[0].m_create(Method)

        def resolve_functional_combination(functionals):
            names = []
            for f_a in functionals:
                name = self._xc_functional_map.get(f_a[0])
                if name is not None:
                    names.append([name, None])
                    if len(f_a) > 1:
                        names[-1][1] = f_a[1]
                for f_b in functionals:
                    if len(f_a) < 3 or len(f_b) < 3:
                        continue
                    combination = '+'.join(['%s_%1.3f_%s' % tuple(f) for f in [f_a, f_b]])
                    name = self._xc_functional_map.get(combination)
                    if name is not None:
                        names.append([name, None])
            return names

        sec_method.electronic_structure_method = 'DFT'

        dft = source.get('dft', [{}])[-1]

        general_info = dft.get('general_info', {})
        for key, val in general_info.get('info', []):
            key = self._metainfo_map.get(key)
            if key is None:
                continue
            if key == 'scf_threshold_energy_change':
                val = fix_dfloat(val) * ureg.hartree
            setattr(sec_method, key, val)

        xc_functionals = dft.get('xc_info', {}).get('functional', [])
        for name, weight in resolve_functional_combination(xc_functionals):
            sec_xc_functional = sec_method.m_create(XCFunctionals)
            sec_xc_functional.XC_functional_name = name
            if weight is not None:
                sec_xc_functional.XC_functional_weight = weight

        parameters = source.get('parameters', [])
        for parameter in parameters:
            key = self._metainfo_map.get(parameter[0])
            if len(parameter) == 2 and key is not None:
                setattr(sec_method, 'x_nwchem_%s' % key, parameter[1])

        total_charge = source.get('total_charge')
        if total_charge is not None:
            sec_method.total_charge = total_charge

        return sec_method

    def parse_configurations(self):

        def parse_calculation(source):
            sec_method = self.parse_method(source)
            sec_system = self.parse_system(source)
            sec_scc = self.parse_scc(source)
            sec_scc.single_configuration_to_calculation_method_ref = sec_method
            sec_scc.single_configuration_calculation_to_system_ref = sec_system

        if self.out_parser.get('geometry_optimization') is not None:
            for iteration in self.out_parser.get('geometry_optimization').get('iteration', []):
                parse_calculation(iteration)

        elif self.out_parser.get('molecular_dynamics') is not None:
            for iteration in self.out_parser.get('molecular_dynamics').get('iteration', []):
                parse_calculation(iteration)

        elif self.out_parser.get('pw') is not None:
            for calculation in self.out_parser.get('pw'):
                parse_calculation(calculation)

        elif self.out_parser.get('single_point') is not None:
            parse_calculation(self.out_parser.get('single_point'))

    def parse_sampling_method(self):
        sec_sampling = self.archive.section_run[0].m_create(SamplingMethod)

        parameters = []
        if self.out_parser.get('geometry_optimization') is not None:
            sec_sampling.sampling_method = 'geometry_optimization'
            parameters = self.out_parser.get('geometry_optimization').get('parameters', [])

        elif self.out_parser.get('molecular_dynamics') is not None:
            sec_sampling.sampling_method = 'molecular_dynamics'
            parameters = self.out_parser.get('molecular_dynamics').get('parameters', [])

        elif self.out_parser.get('pw') is not None:
            # pw is not fully implemented as I do not have example files
            sec_sampling.sampling_method = 'geometry_optimization' if len(
                self.out_parser.get('pw')) > 1 else 'single_point'

        elif self.out_parser.get('single_point') is not None:
            sec_sampling.sampling_method = 'single_point'

        for parameter in parameters:
            if len(parameter) != 2:
                continue
            if parameter[0] == 'maximum gradient threshold         (gmax)':
                val = fix_dfloat(parameter[1]) if isinstance(parameter[1], str) else parameter[1]
                sec_sampling.geometry_optimization_threshold_force = val * ureg.hartree / ureg.bohr
            elif parameter[0] == 'maximum cartesian step threshold   (xmax)':
                val = fix_dfloat(parameter[1]) if isinstance(parameter[1], str) else parameter[1]
                sec_sampling.geometry_optimization_geometry_change = val * ureg.bohr
            elif parameter[0] == 'energy precision                  (eprec)':
                val = fix_dfloat(parameter[1]) if isinstance(parameter[1], str) else parameter[1]
                sec_sampling.geometry_optimization_energy_change = val * ureg.hartree
            else:
                parameter[0] = self._metainfo_map.get(parameter[0])
                if parameter[0] is not None:
                    setattr(sec_sampling, 'x_nwchem_%s' % parameter[0], parameter[1])

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)

        sec_run.program_name = 'NWChem'
        sec_run.program_basis_set_type = 'plane waves' if self.out_parser.get('pw') is not None else 'gaussians'
        sec_run.program_version = self.out_parser.get('version', '')

        # job information
        sec_info = sec_run.m_create(x_nwchem_section_start_information)
        for key, val in self.out_parser.get('job_info', {}).get('info', []):
            if val is None:
                continue
            key = self._metainfo_map.get(key, key.replace(' ', '_'))
            setattr(sec_info, 'x_nwchem_%s' % key, val)

        self.parse_configurations()

        self.parse_sampling_method()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='gulp_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='gulp.nomadmetainfo.json'))


class x_gulp_section_main_keyword(MSection):
    '''
    Section for GULP calculation mode input variable
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gulp_section_main_keyword'))

    x_gulp_main_keyword = Quantity(
        type=str,
        shape=[],
        description='''
        GULP calculation mode input variable
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_main_keyword'))


class x_gulp_section_forcefield(MSection):
    '''
    Section for GULP force field specification
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gulp_section_forcefield'))

    x_gulp_forcefield_species_1 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field species 1
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_species_1'))

    x_gulp_forcefield_species_2 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field species 2
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_species_2'))

    x_gulp_forcefield_species_3 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field species 3
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_species_3'))

    x_gulp_forcefield_species_4 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field species 4
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_species_4'))

    x_gulp_forcefield_speciestype_1 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field speciestype 1
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_speciestype_1'))

    x_gulp_forcefield_speciestype_2 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field speciestype 2
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_speciestype_2'))

    x_gulp_forcefield_speciestype_3 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field speciestype 3
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_speciestype_3'))

    x_gulp_forcefield_speciestype_4 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field speciestype 4
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_speciestype_4'))

    x_gulp_forcefield_potential_name = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field potential name
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_potential_name'))

    x_gulp_forcefield_parameter_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP force field parameter A
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_parameter_a'))

    x_gulp_forcefield_parameter_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP force field parameter B
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_parameter_b'))

    x_gulp_forcefield_parameter_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP force field parameter C
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_parameter_c'))

    x_gulp_forcefield_parameter_d = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP force field parameter D
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_parameter_d'))

    x_gulp_forcefield_cutoff_min = Quantity(
        type=str,
        shape=[],
        description='''
        GULP force field cutoff min (can also be a string like 3Bond for some reason)
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_cutoff_min'))

    x_gulp_forcefield_cutoff_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP force field cutoff max
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_cutoff_max'))

    x_gulp_forcefield_threebody_1 = Quantity(
        type=str,
        shape=[],
        description='''
        GULP 3-body force field parameter 1
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_threebody_1'))

    x_gulp_forcefield_threebody_2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP 3-body force field parameter 2
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_threebody_2'))

    x_gulp_forcefield_threebody_3 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP 3-body force field parameter 3
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_threebody_3'))

    x_gulp_forcefield_threebody_theta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP 3-body force field parameter theta
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_threebody_theta'))

    x_gulp_forcefield_fourbody_force_constant = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP 4-body force field parameter force constant
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_fourbody_force_constant'))

    x_gulp_forcefield_fourbody_sign = Quantity(
        type=str,
        shape=[],
        description='''
        GULP 4-body force field parameter sign
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_fourbody_sign'))

    x_gulp_forcefield_fourbody_phase = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP 4-body force field parameter phase
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_fourbody_phase'))

    x_gulp_forcefield_fourbody_phi0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP 4-body force field parameter phi0
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_forcefield_fourbody_phi0'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_gulp_patterson_group = Quantity(
        type=str,
        shape=[],
        description='''
        Patterson group
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_patterson_group'))

    x_gulp_space_group = Quantity(
        type=str,
        shape=[],
        description='''
        Space group
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_space_group'))

    x_gulp_formula = Quantity(
        type=str,
        shape=[],
        description='''
        GULP chemical formula
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_formula'))

    x_gulp_cell_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_cell_alpha'))

    x_gulp_cell_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_cell_beta'))

    x_gulp_cell_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_cell_gamma'))

    x_gulp_cell_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_cell_a'))

    x_gulp_cell_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_cell_b'))

    x_gulp_cell_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_cell_c'))

    x_gulp_prim_cell_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_prim_cell_alpha'))

    x_gulp_prim_cell_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_prim_cell_beta'))

    x_gulp_prim_cell_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_prim_cell_gamma'))

    x_gulp_prim_cell_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_prim_cell_a'))

    x_gulp_prim_cell_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_prim_cell_b'))

    x_gulp_prim_cell_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_prim_cell_c'))

    x_gulp_pbc = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        grrr
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_pbc'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_gulp_title = Quantity(
        type=str,
        shape=[],
        description='''
        Title of GULP calculation
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_title'))

    x_gulp_section_main_keyword = SubSection(
        sub_section=SectionProxy('x_gulp_section_main_keyword'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gulp_section_main_keyword'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_gulp_number_of_species = Quantity(
        type=int,
        shape=[],
        description='''
        Number of species in GULP
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_number_of_species'))

    x_gulp_species_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['x_gulp_number_of_species'],
        description='''
        Number of species in GULP
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_species_charge'))

    x_gulp_section_forcefield = SubSection(
        sub_section=SectionProxy('x_gulp_section_forcefield'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gulp_section_forcefield'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_gulp_energy_attachment_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for attachment_energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_attachment_energy'))

    x_gulp_energy_attachment_energy_unit = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for attachment_energy_unit
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_attachment_energy_unit'))

    x_gulp_energy_bond_order_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for bond_order_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_bond_order_potentials'))

    x_gulp_energy_brenner_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for brenner_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_brenner_potentials'))

    x_gulp_energy_bulk_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for bulk_energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_bulk_energy'))

    x_gulp_energy_dispersion_real_recip = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for dispersion_real_recip
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_dispersion_real_recip'))

    x_gulp_energy_electric_field_times_distance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for electric_field_times_distance
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_electric_field_times_distance'))

    x_gulp_energy_energy_shift = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for energy_shift
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_energy_shift'))

    x_gulp_energy_four_body_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for four_body_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_four_body_potentials'))

    x_gulp_energy_improper_torsions = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for improper_torsions
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_improper_torsions'))

    x_gulp_energy_interatomic_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for interatomic_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_interatomic_potentials'))

    x_gulp_energy_many_body_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for many_body_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_many_body_potentials'))

    x_gulp_energy_monopole_monopole_real = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for monopole_monopole_real
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_monopole_monopole_real'))

    x_gulp_energy_monopole_monopole_recip = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for monopole_monopole_recip
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_monopole_monopole_recip'))

    x_gulp_energy_monopole_monopole_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for monopole_monopole_total
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_monopole_monopole_total'))

    x_gulp_energy_neutralising_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for neutralising_energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_neutralising_energy'))

    x_gulp_energy_non_primitive_unit_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for non_primitive_unit_cell
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_non_primitive_unit_cell'))

    x_gulp_energy_out_of_plane_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for out_of_plane_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_out_of_plane_potentials'))

    x_gulp_energy_primitive_unit_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for primitive_unit_cell
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_primitive_unit_cell'))

    x_gulp_energy_reaxff_force_field = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for reaxff_force_field
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_reaxff_force_field'))

    x_gulp_energy_region_1_2_interaction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for region_1_2_interaction
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_region_1_2_interaction'))

    x_gulp_energy_region_2_2_interaction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for region_2_2_interaction
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_region_2_2_interaction'))

    x_gulp_energy_self_energy_eem_qeq_sm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for self_energy_eem_qeq_sm
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_self_energy_eem_qeq_sm'))

    x_gulp_energy_sm_coulomb_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for sm_coulomb_correction
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_sm_coulomb_correction'))

    x_gulp_energy_solvation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for solvation_energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_solvation_energy'))

    x_gulp_energy_three_body_potentials = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for three_body_potentials
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_three_body_potentials'))

    x_gulp_energy_total_lattice_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP energy term for total_lattice_energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_energy_total_lattice_energy'))

    x_gulp_md_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP molecular dynamics time
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_md_time'))

    x_gulp_md_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP molecular dynamics kinetic energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_md_kinetic_energy'))

    x_gulp_md_potential_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP molecular dynamics potential energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_md_potential_energy'))

    x_gulp_md_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP molecular dynamics total energy
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_md_total_energy'))

    x_gulp_md_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP molecular dynamics temperature
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_md_temperature'))

    x_gulp_md_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GULP molecular dynamics pressure
        ''',
        a_legacy=LegacyDefinition(name='x_gulp_md_pressure'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask_restplus import Resource, fields
from elasticsearch_dsl import Q

from nomad import search
from nomad.app.flask.api.auth import authenticate

from .api import api, arg_parser, rdf_respose, response_types
from .mapping import Mapping

ns = api.namespace('catalog', description='The API for DCAT catalog.')

iso8601 = fields.DateTime(dt_format='iso8601')

arg_parser = arg_parser.copy()
arg_parser.add_argument('after', type=str)
arg_parser.add_argument(
    'modified_since', type=lambda x: iso8601.parse(x),
    help='A yyyy-MM-ddTHH:mm:ss (RFC3339) maximum entry time (e.g. upload time)')


@ns.route('/')
class Catalog(Resource):
    @api.doc('get_dcat_datasets')
    @api.expect(arg_parser)
    @api.representation('application/xml')
    @api.produces(response_types)
    @api.response(404, 'There is no entry with the given id.')
    @api.response(401, 'This entry is not publically accessible.')
    @api.response(200, 'Data send', headers={'Content-Type': 'application/xml'})
    @authenticate()
    def get(self):
        ''' Returns a page of DCAT datasets. '''
        args = arg_parser.parse_args()
        modified_since = args.get('modified_since', None)
        after = args.get('after', '')
        if after is None:
            after = ''

        search_request = search.SearchRequest().owner('public')
        if modified_since is not None:
            modified_clause = Q('range', upload_time=dict(gte=modified_since))
            modified_clause |= Q('range', last_edit=dict(gte=modified_since))
            modified_clause |= Q('range', last_processing=dict(gte=modified_since))
            search_request.q &= modified_clause

        es_search = search_request._search.query(search_request.q)
        if after != '':
            es_search = es_search.extra(search_after=[after], sort='calc_id')
        es_response = es_search.execute()

        mapping = Mapping()
        mapping.map_catalog(es_response.hits, after, modified_since, slim=False)
        return rdf_respose(mapping.g)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os.path

from nomad import config, datamodel

from .parser import MissingParser, BrokenParser, Parser, ArchiveParser
from .artificial import EmptyParser, GenerateRandomParser, TemplateParser, ChaosParser

from eelsdbconverter import EELSApiJsonConverter
from mpesparser import MPESParser
from aptfimparser import APTFIMParser
from vaspparser import VASPParser
from phonopyparser import PhonopyParser
from elasticparser import ElasticParser
from lammpsparser import LammpsParser
from gromacsparser import GromacsParser
from crystalparser import CrystalParser
from fhiaimsparser import FHIAimsParser
from excitingparser import ExcitingParser
from abinitparser import AbinitParser
from quantumespressoparser import QuantumEspressoParser
from gaussianparser import GaussianParser
from gpawparser import GPAWParser
from octopusparser import OctopusParser
from orcaparser import OrcaParser
from cp2kparser import CP2KParser
from fhivibesparser import FHIVibesParser
from turbomoleparser import TurbomoleParser
from castepparser import CastepParser
from wien2kparser import Wien2kParser
from nwchemparser import NWChemParser
from lobsterparser import LobsterParser
from bandparser import BandParser
from amberparser import AmberParser
from asapparser import AsapParser
from bigdftparser import BigDFTParser
from cpmdparser import CPMDParser
from dftbplusparser import DFTBPlusParser
from dlpolyparser import DLPolyParser
from dmol3parser import Dmol3Parser
from elkparser import ElkParser
from fleurparser import FleurParser
from fploparser import FploParser
from gamessparser import GamessParser
from gromosparser import GromosParser
from gulpparser import GulpParser
from molcasparser import MolcasParser
from mopacparser import MopacParser
from namdparser import NAMDParser
from onetepparser import OnetepParser
from siestaparser import SiestaParser
from tinkerparser import TinkerParser
from charmmparser import CharmmParser
from libatomsparser import LibAtomsParser
from atkparser import ATKParser
from qboxparser import QboxParser
from openkimparser import OpenKIMParser
from openmxparser import OpenmxParser

try:
    # these packages are not available without parsing extra, which is ok, if the
    # parsers are only initialized to load their metainfo definitions
    import magic
    import gzip
    import bz2
    import lzma

    _compressions = {
        b'\x1f\x8b\x08': ('gz', gzip.open),
        b'\x42\x5a\x68': ('bz2', bz2.open),
        b'\xfd\x37\x7a': ('xz', lzma.open)
    }

    encoding_magic = magic.Magic(mime_encoding=True)

except ImportError:
    pass


def match_parser(mainfile_path: str, strict=True) -> Parser:
    '''
    Performs parser matching. This means it take the given mainfile and potentially
    opens it with the given callback and tries to identify a parser that can parse
    the file.

    This is determined by filename (e.g. *.out), mime type (e.g. text/*, application/xml),
    and beginning file contents.

    Arguments:
        mainfile_path: Path to the mainfile
        strict: Only match strict parsers, e.g. no artificial parsers for missing or empty entries.

    Returns: The parser, or None if no parser could be matched.
    '''
    mainfile = os.path.basename(mainfile_path)
    if mainfile.startswith('.') or mainfile.startswith('~'):
        return None

    with open(mainfile_path, 'rb') as f:
        compression, open_compressed = _compressions.get(f.read(3), (None, open))

    with open_compressed(mainfile_path, 'rb') as cf:  # type: ignore
        buffer = cf.read(config.parser_matching_size)

    mime_type = magic.from_buffer(buffer, mime=True)

    decoded_buffer = None
    encoding = None
    try:  # Try to open the file as a string for regex matching.
        decoded_buffer = buffer.decode('utf-8')
    except UnicodeDecodeError:
        # This file is either binary or has wrong encoding
        encoding = encoding_magic.from_buffer(buffer)

        if config.services.force_raw_file_decoding:
            encoding = 'iso-8859-1'

        if encoding in ['iso-8859-1']:
            try:
                decoded_buffer = buffer.decode(encoding)
            except Exception:
                pass

    for parser in parsers:
        if strict and isinstance(parser, (MissingParser, EmptyParser)):
            continue

        if parser.is_mainfile(mainfile_path, mime_type, buffer, decoded_buffer, compression):
            # potentially convert the file
            if encoding in ['iso-8859-1']:
                try:
                    with open(mainfile_path, 'rb') as binary_file:
                        content = binary_file.read().decode(encoding)
                except Exception:
                    pass
                else:
                    with open(mainfile_path, 'wt') as text_file:
                        text_file.write(content)

            # TODO: deal with multiple possible parser specs
            return parser

    return None


parsers = [
    GenerateRandomParser(),
    TemplateParser(),
    ChaosParser(),
    PhonopyParser(),
    VASPParser(),
    ExcitingParser(),
    FHIAimsParser(),
    FHIVibesParser(),
    CP2KParser(),
    CrystalParser(),
    CPMDParser(),
    NWChemParser(),
    BigDFTParser(),
    Wien2kParser(),
    BandParser(),
    QuantumEspressoParser(),
    GaussianParser(),
    AbinitParser(),
    OrcaParser(),
    CastepParser(),
    DLPolyParser(),
    LibAtomsParser(),
    OctopusParser(),
    GPAWParser(),
    ATKParser(),
    GulpParser(),
    SiestaParser(),
    ElkParser(),
    ElasticParser(),
    GamessParser(),
    TurbomoleParser(),
    MPESParser(),
    APTFIMParser(),
    EELSApiJsonConverter(),
    QboxParser(),
    Dmol3Parser(),
    FleurParser(),
    MolcasParser(),
    OnetepParser(),
    OpenKIMParser(),
    TinkerParser(),
    LammpsParser(),
    AmberParser(),
    GromacsParser(),
    LobsterParser(),
    GromosParser(),
    NAMDParser(),
    CharmmParser(),
    DFTBPlusParser(),
    AsapParser(),
    FploParser(),
    MopacParser(),
    OpenmxParser(),
    ArchiveParser()
]

empty_parsers = [
    EmptyParser(
        name='missing/octopus', code_name='Octopus', code_homepage='https://octopus-code.org/',
        domain='dft',
        mainfile_name_re=r'(inp)|(.*/inp)'
    ),
    EmptyParser(
        name='missing/crystal', code_name='Crystal', code_homepage='https://www.crystal.unito.it/index.php',
        domain='dft',
        mainfile_name_re=r'.*\.cryst\.out'
    ),
    EmptyParser(
        name='missing/wien2k', code_name='WIEN2k', code_homepage='http://www.wien2k.at/',
        domain='dft',
        mainfile_name_re=r'.*\.scf'
    ),
    EmptyParser(
        name='missing/fhi-aims', code_name='FHI-aims', code_homepage='https://aimsclub.fhi-berlin.mpg.de/',
        domain='dft',
        mainfile_name_re=r'.*\.fhiaims'
    )
]

if config.use_empty_parsers:
    # There are some entries with PIDs that have mainfiles which do not match what
    # the actual parsers expect. We use the EmptyParser to produce placeholder entries
    # to keep the PIDs. These parsers will not match for new, non migrated data.
    parsers.extend(empty_parsers)

parsers.append(BrokenParser())

''' Instantiation and constructor based config of all parsers. '''

parser_dict = {parser.name: parser for parser in parsers + empty_parsers}  # type: ignore
''' A dict to access parsers by name. Usually 'parsers/<...>', e.g. 'parsers/vasp'. '''

# renamed parsers
parser_dict['parser/broken'] = parser_dict['parsers/broken']
parser_dict['parser/fleur'] = parser_dict['parsers/fleur']
parser_dict['parser/molcas'] = parser_dict['parsers/molcas']
parser_dict['parser/octopus'] = parser_dict['parsers/octopus']
parser_dict['parser/onetep'] = parser_dict['parsers/onetep']

# register code names as possible statistic value to the dft datamodel
code_names = []
for parser in parsers:
    if parser.domain == 'dft' and \
            getattr(parser, 'code_name', None) is not None and \
            getattr(parser, 'code_name') != 'currupted mainfile' and \
            getattr(parser, 'code_name') != 'Template':
        code_names.append(getattr(parser, 'code_name'))
code_names = sorted(set(code_names), key=lambda code_name: code_name.lower())
datamodel.DFTMetadata.code_name.a_search.statistic_values = code_names + [
    config.services.unavailable_value, config.services.not_processed_value]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from abinitparser.metainfo import abinit_autogenerated
from nomad.datamodel.metainfo import public

m_package = Package(
    name='abinit_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='abinit.nomadmetainfo.json'))


class x_abinit_section_stress_tensor(MSection):
    '''
    Section describing the stress tensor
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_abinit_section_stress_tensor'))

    x_abinit_stress_tensor_xx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        xx component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_xx'))

    x_abinit_stress_tensor_yy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        yy component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_yy'))

    x_abinit_stress_tensor_zz = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        zz component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_zz'))

    x_abinit_stress_tensor_zy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        zy component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_zy'))

    x_abinit_stress_tensor_zx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        zx component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_zx'))

    x_abinit_stress_tensor_yx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        yx component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_yx'))


class x_abinit_section_dataset_header(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_abinit_section_dataset_header'))

    x_abinit_dataset_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Dataset number
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_dataset_number'))

    x_abinit_vprim_1 = Quantity(
        type=str,
        shape=[],
        description='''
        Primitive axis 1
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vprim_1'))

    x_abinit_vprim_2 = Quantity(
        type=str,
        shape=[],
        description='''
        Primitive axis 2
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vprim_2'))

    x_abinit_vprim_3 = Quantity(
        type=str,
        shape=[],
        description='''
        Primitive axis 3
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vprim_3'))


class x_abinit_section_var(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_abinit_section_var'))

    x_abinit_vardtset = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Variable dataset number
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vardtset'))

    x_abinit_varname = Quantity(
        type=str,
        shape=[],
        description='''
        Variable name
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_varname'))

    x_abinit_varvalue = Quantity(
        type=str,
        shape=[],
        description='''
        Variable value
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_varvalue'))

    x_abinit_vartruncation = Quantity(
        type=str,
        shape=[],
        description='''
        Variable truncation length
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vartruncation'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_abinit_parallel_compilation = Quantity(
        type=str,
        shape=[],
        description='''
        Parallel or sequential compilation
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_parallel_compilation'))

    x_abinit_start_date = Quantity(
        type=str,
        shape=[],
        description='''
        Start date as string
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_start_date'))

    x_abinit_start_time = Quantity(
        type=str,
        shape=[],
        description='''
        Start time as string
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_start_time'))

    x_abinit_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        Input file name
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_input_file'))

    x_abinit_output_file = Quantity(
        type=str,
        shape=[],
        description='''
        Output file name
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_output_file'))

    x_abinit_input_files_root = Quantity(
        type=str,
        shape=[],
        description='''
        Root for input files
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_input_files_root'))

    x_abinit_output_files_root = Quantity(
        type=str,
        shape=[],
        description='''
        Root for output files
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_output_files_root'))

    x_abinit_total_cpu_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total CPU time
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_total_cpu_time'))

    x_abinit_total_wallclock_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total wallclock time
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_total_wallclock_time'))

    x_abinit_completed = Quantity(
        type=str,
        shape=[],
        description='''
        Message that the calculation was completed
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_completed'))

    x_abinit_section_var = SubSection(
        sub_section=SectionProxy('x_abinit_section_var'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_abinit_section_var'))


class section_method(public.section_method):
    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_abinit_tolvrs = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        `TOLerance on the potential V(r) ReSidual`:
        Sets a tolerance for potential residual that, when reached, will cause
        one SCF cycle to stop (and ions to be moved). If set to zero, this
        stopping condition is ignored. Instead, refer to other tolerances, such
        as toldfe, tolwfr.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_tolvrs'))

    x_abinit_tolwfr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOLerance on WaveFunction squared Residual:
        Specifies the threshold on WaveFunction squared Residuals;
        it gives a convergence tolerance for the largest squared residual
        for any given band.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_tolwfr'))

    x_abinit_istwfk = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
         Integer for choice of STorage of WaveFunction at each k point;
        Controls the way the wavefunction for each k-point is stored inside ABINIT,
        in reciprocal space, according to time-reversal symmetry properties.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_istwfk'))

    x_abinit_iscf = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ABINIT variable Integer for Self-Consistent-Field cycles
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_iscf'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_abinit_atom_xcart_final = Quantity(
        type=str,
        shape=[],
        description='''
        Cartesian coordinates of an atom at the end of the dataset
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_xcart_final'))

    x_abinit_atom_xcart = Quantity(
        type=str,
        shape=[],
        description='''
        Cartesian coordinates of an atom at the end of a single configuration calculation
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_xcart'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_abinit_magnetisation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total magnetisation.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_magnetisation'))

    x_abinit_fermi_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Fermi energy.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_fermi_energy'))

    x_abinit_single_configuration_calculation_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Determines whether a single configuration calculation is converged.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_single_configuration_calculation_converged'))

    x_abinit_atom_force = Quantity(
        type=str,
        shape=[],
        description='''
        Force acting on an atom at the end of a single configuration calculation
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_force'))

    x_abinit_atom_force_final = Quantity(
        type=np.dtype(np.float64),
        unit='newton',
        shape=[],
        description='''
        Force acting on an atom at the end of the dataset
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_force_final'))

    x_abinit_energy_ewald = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Ewald energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_ewald'))

    x_abinit_energy_psp_core = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pseudopotential core energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_psp_core'))

    x_abinit_energy_psp_local = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Local pseudopotential energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_psp_local'))

    x_abinit_energy_psp_nonlocal = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Non-local pseudopotential energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_psp_nonlocal'))

    x_abinit_energy_internal = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Internal energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_internal'))

    x_abinit_energy_ktentropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -kT*entropy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_ktentropy'))

    x_abinit_energy_band = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Band energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_band'))

    x_abinit_section_stress_tensor = SubSection(
        sub_section=SectionProxy('x_abinit_section_stress_tensor'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_abinit_section_stress_tensor'))

    x_abinit_unit_cell_volume = Quantity(
        type=np.dtype(np.float64),
        unit='meter**3',
        shape=[],
        description='''
        Unit cell volume
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_unit_cell_volume'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    x_abinit_kpt = Quantity(
        type=str,
        shape=[],
        description='''
        The reduced coordinates of a k-point
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_kpt'))

    x_abinit_wtk = Quantity(
        type=str,
        shape=[],
        description='''
        A k-point weight
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_wtk'))

    x_abinit_eigenvalues = Quantity(
        type=str,
        shape=[],
        description='''
        List of eigenvalues on a given k-point
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_eigenvalues'))

    x_abinit_occupations = Quantity(
        type=str,
        shape=[],
        description='''
        List of occupations on a given k-point
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_occupations'))


class x_abinit_section_dataset(abinit_autogenerated.x_abinit_section_dataset):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='x_abinit_section_dataset'))

    x_abinit_geometry_optimization_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Determines whether a geometry optimization is converged.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_geometry_optimization_converged'))

    x_abinit_eig_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Name of file where the eigenvalues were written to.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_eig_filename'))

    x_abinit_section_dataset_header = SubSection(
        sub_section=SectionProxy('x_abinit_section_dataset_header'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_abinit_section_dataset_header'))


m_package.__init_metainfo__()



from .file_parser import FileParser
from .text_parser import TextParser, DataTextParser, Quantity, ParsePattern
from .xml_parser import XMLParser
from .tar_parser import TarParser
from .basic_parser import BasicParser

UnstructuredTextFileParser = TextParser



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
A command that generates various statistics.
'''

import matplotlib.scale as mscale
import matplotlib.transforms as mtransforms
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import click
import json
import datetime
import subprocess

from nomad import config
from .client import client


def codes(client, minimum=1, **kwargs):
    data = client.repo.search(per_page=1, **kwargs).response().result

    x_values = sorted([
        code for code, values in data.statistics['dft.code_name'].items()
        if code != 'not processed' and values.get('calculations', 1000) >= minimum], key=lambda x: x.lower())

    return data.statistics, x_values, 'dft.code_name', 'code'


def dates(client, minimum=1, **kwargs):
    data = client.repo.search(per_page=1, date_histogram=True, **kwargs).response().result

    x_values = list([
        x for x in data.statistics['date_histogram'].keys()])

    return data.statistics, x_values, 'date_histogram', 'month'


def error_fig(client):
    _, labels, _, _ = codes(client)

    def code_values(metric='code_runs', **kwargs):
        result = client.repo.search(
            per_page=1,
            owner='admin',
            metrics=[] if metric == 'code_runs' else metric,
            **kwargs).response().result

        return {
            code: values[metric]
            for code, values in result.quantities['dft.code_name'].items()
            if code != 'not processed' and (not labels or code in labels) > 0}

    # get the data
    all_entries = code_values()
    parser_failure_label = 'parser failure'
    error_types = [
        {'name': parser_failure_label, 'search': dict(system='not processed')},
        {'name': 'failed system classification', 'search': dict(system='unavailable')},
        {'name': 'no basis set available', 'search': dict(basis_set='unavailable')},
        {'name': 'no XC functional available', 'search': dict(xc_functional='unavailable')}
    ]
    errors = {
        error_type['name']: {
            code: failures
            for code, failures in code_values(**error_type['search']).items()}
        for error_type in error_types}
    errors_rates = {
        error_type['name']: {
            code: 0 if all_entries[code] == 0 else failures / all_entries[code]
            for code, failures in code_values(**error_type['search']).items()}
        for error_type in error_types}

    fig, axs = plt.subplots(figsize=(15, 12), dpi=72, nrows=2)

    def draw_error_chart(errors, ax, colors, entries=None, mul=1, scale=0.5):
        n_bars = len(errors) - 1
        leg_colors = list(colors)

        x = np.arange(len(labels))  # the label locations
        width = 0.7 / n_bars  # the width of the bars
        plt.sca(ax)
        plt.xticks(rotation=90)

        if entries is not None:
            ax.bar(x, [entries[code] for code in labels], width * n_bars, label='all entries', color=colors.pop(0))

        i = -1
        not_processed = [errors[parser_failure_label][code] * mul for code in labels]
        ax.bar(x, not_processed, width * n_bars, label=parser_failure_label, color=colors.pop(0))
        for key, values in errors.items():
            if key != parser_failure_label:
                ax.bar(x + i * width, [values[code] * mul for code in labels], width, label=key, bottom=not_processed, color=colors.pop(0))
                i += 1

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_yscale('power', exponent=scale)
        ax.set_xticks(x)
        ax.set_xticklabels(labels)
        ax.legend()
        leg = ax.get_legend()
        for i in range(0, len(leg_colors)):
            leg.legendHandles[i].set_color(leg_colors[i])

        fig.tight_layout()

    ax = axs[0]
    ax.set_title('Absolute number of entries with parser errors or missing repository metadata compared to all entries per code')
    ax.set_ylabel('number of entries', )
    colors = ['grey', 'red', 'yellow', 'orange', 'brown']
    draw_error_chart(errors, ax, entries=all_entries, mul=1, scale=0.25, colors=colors)
    ax.yaxis.set_major_formatter(ticker.StrMethodFormatter('{x:,.0f}'))

    ax = axs[1]
    ax.set_title('Relative rates of entries with parser errors or missing repository metadata per code')
    ax.set_ylabel('rate in %', )
    colors = ['red', 'yellow', 'orange', 'brown']
    draw_error_chart(errors_rates, ax, mul=100, colors=colors)

    plt.show()

    return fig, plt


class Metric:
    def __init__(self, metric, label=None, power=None, multiplier=1, format=None, cumulate=False):
        if label is None:
            label = metric

        self.metric = metric
        self.agg = None
        self.label = label
        self.multiplier = multiplier
        self.power = power
        self.format = format
        self.cumulate = cumulate

    def draw_axis(self, axis, data, x_values, x_positions, width, color, only=False):
        label_color = 'black' if only else color
        value_map = {
            x: values[self.metric]
            for x, values in data[self.agg].items()
            if x in x_values}

        if self.power is not None:
            axis.set_yscale('power', exponent=self.power)
        else:
            axis.set_yscale('log')
        axis.set_ylabel(self.label, color=label_color)

        if self.format is not None:
            axis.yaxis.set_major_formatter(ticker.StrMethodFormatter(self.format))

        y_values = [value_map[x] * self.multiplier for x in x_values]
        if self.cumulate:
            y_values = np.array(y_values).cumsum()
        axis.bar(x_positions, y_values, width, label=self.label, color=color, align='edge')
        axis.tick_params(axis='y', labelcolor=label_color)

        # TODO remove
        # for x, v in zip(x_positions, y_values):
        #     axis.text(x + .1, v, ' {:,}'.format(int(v)), color=color, fontweight='bold', rotation=90)
        # import matplotlib.lines as mlines
        # line = mlines.Line2D([min(x_positions), max(x_positions)], [72, 72], color=color)
        # axis.add_line(line)


def bar_plot(
        client, retrieve, metric1, metric2=None, title=None, format_xlabel=None,
        xlim={}, ylim=dict(bottom=1), **kwargs):
    if format_xlabel is None:
        format_xlabel = lambda x: x

    metrics = [] if metric1.metric == 'code_runs' else [metric1.metric]
    if metric2 is not None:
        metrics += [] if metric2.metric == 'code_runs' else [metric2.metric]

    data, x_values, agg, agg_label = retrieve(client, metrics=metrics, statistics=['dft.code_name'], **kwargs)
    metric1.agg = agg
    if metric2 is not None:
        metric2.agg = agg

    fig, ax1 = plt.subplots(figsize=(7, 4), dpi=72)
    x = np.arange(len(x_values))
    width = 0.8 / 2
    if metric2 is None:
        width = 0.8
    plt.sca(ax1)
    plt.xticks(rotation=90)
    ax1.set_xticks(x)
    ax1.set_xticklabels([format_xlabel(value) if value != 'Quantum Espresso' else 'Q. Espresso' for value in x_values])
    ax1.margins(x=0.01)
    # ax1.set_xlim(**xlim)
    # i = 0
    # for label in ax1.xaxis.get_ticklabels():
    #     label.set_visible(i % 4 == 0)
    #     i += 1

    if title is None:
        title = 'Number of %s' % metric1.label
        if metric2 is not None:
            title += ' and %s' % metric2.label
        title += ' per %s' % agg_label
        ax1.set_title(title)
    elif title != '':
        ax1.set_title(title)

    metric1.draw_axis(ax1, data, x_values, x - (width / 2), width, 'tab:blue', only=metric2 is None)
    ax1.set_ylim(**ylim)
    # ax1.set_yticks([40, 30, 20, 10, 5, 1, 0.5, 0.1])
    if not metric2:
        ax1.grid(which='major', axis='y', linestyle='-')

    if metric2:
        ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
        metric2.draw_axis(
            ax2, data, x_values, x + width / 2, width, 'tab:red')
        ax2.set_ylim(bottom=1)

    fig.tight_layout()

    return fig, plt


@client.command(help='Generate various matplotlib charts')
@click.option('--errors', is_flag=True, help='Two charts with relative and absolute parser/normalizer errors per code.')
@click.option('--x-axis', type=str, help='Aggregation used for x-axis, values are "code" and "time".')
@click.option('--y-axis', multiple=True, type=str, help='Metrics used for y-axis, values are "entries", "energies", "calculations", "uploaders".')
@click.option('--cumulate', is_flag=True, help='Cumulate over x-axis.')
@click.option('--title', type=str, help='Override chart title with given value.')
@click.option('--total', is_flag=True, help='Provide total sums of key metrics.')
@click.option('--save', type=str, help='Save to given file instead of showing the plot.')
@click.option('--power', type=float, help='User power scale instead of log with the given inverse power.')
@click.option('--open-access', is_flag=True, help='Only consider Open-Access data.')
@click.option('--minimum', type=int, default=1, help='Only consider codes with at least the given ammount of entries.')
def statistics_plot(errors, title, x_axis, y_axis, cumulate, total, save, power, open_access, minimum):
    from .client import create_client
    client = create_client()

    class PowerScale(mscale.ScaleBase):
        name = 'power'

        def __init__(self, axis, exponent, **kwargs):
            mscale.ScaleBase.__init__(self, axis, **kwargs)
            self.exponent = exponent

        def set_default_locators_and_formatters(self, axis):
            axis.set_major_locator(ticker.AutoLocator())
            axis.set_major_formatter(ticker.ScalarFormatter())
            axis.set_minor_locator(ticker.NullLocator())
            axis.set_minor_formatter(ticker.NullFormatter())

        def limit_range_for_scale(self, vmin, vmax, minpos):
            return max(0., vmin), vmax

        class Transform(mtransforms.Transform):
            input_dims = 1
            output_dims = 1
            is_separable = True

            def __init__(self, exponent):
                super().__init__()
                self.exponent = exponent

            def transform_non_affine(self, a):
                return np.array(a)**self.exponent

            def inverted(self):
                return PowerScale.Transform(1 / self.exponent)

        def get_transform(self):
            return self.Transform(self.exponent)

    mscale.register_scale(PowerScale)

    kwargs = {}
    if cumulate:
        kwargs.update(
            power=1,
            multiplier=1e-6,
            format='{x:,.1f}M')
    elif power is not None:
        kwargs.update(
            power=1 / power,
            multiplier=1e-6,
            format='{x:,.1f}M')

    metrics = {
        'entries': Metric(
            'code_runs',
            label='entries (code runs)',
            cumulate=cumulate,
            **kwargs),
        'uploaders': Metric(
            'uploaders',
            cumulate=cumulate,
            label='uploaders that provided data'),
        'energies': Metric(
            'dft.total_energies',
            label='total energy calculations',
            cumulate=cumulate,
            **kwargs),
        'calculations': Metric(
            'dft.calculations',
            label='calculations (e.g. total energy)',
            cumulate=cumulate,
            **kwargs)
    }

    if errors:
        fig, plt = error_fig(client)

    owner = 'all' if open_access else 'admin'

    if x_axis is not None:
        assert 1 <= len(y_axis) <= 2, 'Need 1 or 2 y axis'

        kwargs = {}
        if x_axis == 'code':
            x_axis = codes
            kwargs.update(ylim=dict(bottom=1))
        elif x_axis == 'time':
            x_axis = dates
            kwargs.update(
                ylim=dict(bottom=1),
                format_xlabel=lambda x: datetime.datetime.fromtimestamp(int(x) / 1000).strftime('%b %y'))
        else:
            assert False, 'x axis can only be "code" or "time"'

        y_axis = [metrics[y] for y in y_axis]

        fig, plt = bar_plot(
            client, x_axis, *y_axis, title=title, owner=owner, minimum=minimum, **kwargs)

    if errors or x_axis is not None:
        if save is not None:
            fig.savefig(save, bbox_inches='tight')
        else:
            plt.show()

    if total:
        data = client.repo.search(
            per_page=1, owner=owner, statistics=True,
            metrics=['total_energies', 'calculations', 'uploaders', 'authors', 'datasets']).response().result
        print(json.dumps(data.statistics['total'], indent=4))


@client.command(help='Generate table with basic statistics summary.')
@click.option('--html', is_flag=True, help='Output HTML instead of plain text table.')
@click.option('--geometries', is_flag=True, help='Use geometries not unique geometries.')
@click.option('--public-path', type=str, default=config.fs.public, help='The path to the public data. Default is %s.' % config.fs.public)
def statistics_table(html, geometries, public_path):
    # get more stats for files
    # uploads: find . -maxdepth 2 | wc -l
    # public archive: find . -regex '.*archive.*public.*zip' -type f -print0 | du --files0-from=- -ch | grep total$
    # public raw: find . -regex '.*raw.*public.*zip' -type f -print0 | du --files0-from=- -ch | grep total$

    def get_statistic(response, quantity, value, metric):
        quantity_data = response.statistics.get(quantity)
        if quantity_data is None:
            return 0
        value_data = quantity_data.get(value)
        if value_data is None:
            return 0

        value = value_data.get(metric)
        return value if value is not None else 0

    from nomad.cli.client import create_client
    client = create_client()

    geometry_metric = 'dft.unique_geometries' if not geometries else 'dft.geometries'

    # search scc with system type
    data_all = client.repo.search(
        per_page=1, metrics=['dft.calculations'], statistics=['dft.system', 'dft.code_name']).response().result

    entries = get_statistic(data_all, 'total', 'all', 'code_runs')
    calculations = get_statistic(data_all, 'total', 'all', 'dft.calculations')
    calculations_1d = get_statistic(data_all, 'dft.system', '1D', 'dft.calculations') \
        + get_statistic(data_all, 'dft.system', 'atom', 'dft.calculations') \
        + get_statistic(data_all, 'dft.system', 'molecule / cluster', 'dft.calculations')

    calculations_2d = get_statistic(data_all, 'dft.system', '2D / surface', 'dft.calculations')
    calculations_2d += get_statistic(data_all, 'dft.system', '2D', 'dft.calculations')
    calculations_2d += get_statistic(data_all, 'dft.system', 'surface', 'dft.calculations')
    calculations_3d = get_statistic(data_all, 'dft.system', 'bulk', 'dft.calculations')

    metrics_all = client.repo.search(per_page=1, metrics=[geometry_metric, 'dft.quantities']).response().result
    geometries = get_statistic(metrics_all, 'total', 'all', geometry_metric)
    quantities = get_statistic(metrics_all, 'total', 'all', 'dft.quantities')

    # search calcs quantities=section_k_band
    band_structures = get_statistic(
        client.repo.search(per_page=1, **{'dft.quantities': ['section_k_band']}).response().result,
        'total', 'all', 'code_runs')

    # search calcs quantities=section_dos
    dos = get_statistic(
        client.repo.search(per_page=1, **{'dft.quantities': ['section_dos']}).response().result,
        'total', 'all', 'code_runs')

    phonons = get_statistic(
        client.repo.search(per_page=1, **{'dft.code_name': 'Phonopy'}).response().result,
        'total', 'all', 'code_runs')

    # files and sized
    def run_shell_command(command):
        process = subprocess.run(['bash', '-c', command], stdout=subprocess.PIPE)
        out = process.stdout.decode('utf-8')
        if process.stderr is not None:
            err = process.stderr.decode('utf-8')
            print('There is an error: %s' % str(err.strip()))
        return out.split('\t')[0].strip()

    archive_data = run_shell_command((
        'find %s -regex \'.*archive.*public.*zip\' '
        '-type f -print0 | du --files0-from=- -ch | grep total$') % public_path)
    raw_data = run_shell_command((
        'find %s -regex \'.*raw.*public.*zip\' '
        '-type f -print0 | du --files0-from=- -ch | grep total$') % public_path)
    n_uploads = run_shell_command(
        'find %s -regex \'.*raw.*public.*zip\' -type f | wc -l' % public_path)

    try:
        n_uploads = '{:,}'.format(int(n_uploads))
    except Exception:
        pass

    if not html:
        print('''
            Entries: {:,.0f}
            Calculations, e.g. total energies: {:,.0f}
            Geometries: {:,.0f}
            Bulk crystals: {:,.0f}
            2D / Surfaces: {:,.0f}
            Atoms / Molecules: {:,.0f}
            DOS: {:,.0f}
            Band structures: {:,.0f}
            Total parsed quantities: {:,.0f}
            Public raw data: {}B
            Public archive data: {}B
            Number of uploads: {}
        '''.format(
            entries,
            calculations,
            geometries,
            calculations_3d,
            calculations_2d,
            calculations_1d,
            dos,
            band_structures,
            quantities,
            raw_data,
            archive_data,
            n_uploads
        ))

    else:
        print('''
            <div class="container">
                <p>The <i>NOMAD Archive</i> stores calculations performed
                with all the most important and widely used electronic-structure and force-field codes
                in a code-independent format.
                </p>
                <p>Summary statistics of the Archive content (last update in {}):</p>
                <table class="table" style="text-align: left; max-width: 700px;">
                    <thead>
                        <tr>
                        <th scope="col">Metric</th>
                        <th scope="col">Value</th>
                        </tr>
                    </thead>
                    <tbody>
                        <tr>
                        <td scope="row">Entries, i.e. code runs</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Calculations, e.g. total energies</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Geometries</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Bulk Crystals</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Surfaces</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Molecules/Clusters</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">DOS</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Band Structures</td>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Phonon Calculations</th>
                        <td>{:,.0f}</td>
                        </tr>
                        <tr>
                        <td scope="row">Overall parsed quantities</td>
                        <td>{:,.0f}</td>
                        </tr>
                    </tbody>
                </table>
                <p>
                    Furthermore:
                </p>
                <ul>
                    <li><b>{}</b> Uploads with <b>{}B</b> of raw data</li>
                    <li><b>{}B</b> of archive data</li>
                    <li>Data classified using <b>168</b> public metadata of the NOMAD Meta Info and <b>2,360</b> code-specific metadata</li>
                </ul>
                <p>
                    For more and interactive statistics, use the <i>metadata</i> view of
                    the <a href="https://nomad-lab.eu/prod/rae/gui/search">NOMAD Repository and Archvi search</a>.
                </p>
                <p>
                    90% of VASP calculations are provided by
                        <a href="http://aflowlib.org">AFLOWlib</a> (S. Curtarolo),
                        <a href="http://oqmd.org"> OQMD</a> (C. Wolverton) and
                        <a href="https://materialsproject.org">Materials Project</a> (K. Persson).
                </p>
                <p>
                    You can further explore the statistics in the below dynamic histograms. To
                    change the displayed quantity, select from the "Quantities" drop-down. To
                    filter the data, click histogram bars for different filter combinations.
                    To reset filters, click "Reset Filters".
                </p>
                <p>
                    The archive data is represented in a code-independent, structured
                    form. The archive structure and all quantities are described via the
                    <a href="https://nomad-lab.eu/prod/rae/gui/metainfo">NOMAD Metainfo</a>.
                    The NOMAD Metainfo defines a conceptual model to store the values connected
                    to atomistic or <i>ab initio</i> calculations. A clear and usable metadata definition
                    is a prerequisites to preparing the data for analysis that everybody
                    can contribute to.
                </p>
                <p>
                    In collaboration with the <a href="http://www.bbdc.berlin/">Berlin Big Data Center (BBDC)</a>,
                    we use the Apache Flink infrastructure to support and go beyond the standard MapReduce model to enable
                    rapid and complex queries.
                </p>
                <p>
                    Contact concerning general aspects of the CoE: <a href="mailto:pietsch@fhi-berlin.mpg.de">Jessica Pietsch</a>
                </p>
                <p>
                    Contact concerning the NOMAD Archive:
                        <a href="mailto:markus.scheidgen@physik.hu-berlin.de">Markus Scheidgen</a>,
                        <a href="mailto:ghiringhelli@fhi-berlin.mpg.de">Luca Ghiringhelli</a>
                </p>
            </div>
        '''.format(
            datetime.datetime.now().strftime('%b %Y'),
            entries,
            calculations,
            geometries,
            calculations_3d,
            calculations_2d,
            calculations_1d,
            dos,
            band_structures,
            phonons,
            quantities,
            n_uploads,
            raw_data,
            archive_data
        ))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class SiestaParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        super().__init__(
            specifications=dict(
                name='parsers/siesta', code_name='Siesta', code_homepage='https://departments.icmab.es/leem/siesta/',
                mainfile_contents_re=(
                    r'(Siesta Version: siesta-|SIESTA [0-9]\.[0-9]\.[0-9])|'
                    r'(\*\s*WELCOME TO SIESTA\s*\*)')),
            units_mapping=dict(energy=ureg.eV, length=ureg.angstrom),
            program_version=r'Siesta Version\: (siesta\S+)',
            lattice_vectors=r'outcell\: Unit cell vectors \(Ang\)\:([\s\S]+?)\n *\n',
            # TODO support for various input, output formats
            atom_labels_atom_positions_scaled=r'outcoor\: Atomic coordinates \((?:fractional|scaled)\)\:([\s\S]+?\n *\n)',
            atom_labels_atom_positions=r'outcoor\: Atomic coordinates \(Ang\)\:([\s\S]+?\n *\n)',
            energy_method_current=rf'siesta\: E_KS\(eV\) = +({re_f})',
            energy_total=rf'siesta\: Etot +\= +({re_f})',
            atom_forces=r'siesta: Atomic forces \(eV/Ang\)\:([\s\S]+?)\-{10}')



from typing import Callable
from lark import v_args
from elasticsearch_dsl import Q, Field, Text, Keyword, Integer, Boolean
from optimade.filtertransformers.elasticsearch import Quantity
from nomad.atomutils import get_hill_decomposition

from nomad.app.optimade.filterparser import ElasticTransformer


_cmp_operators = {">": "gt", ">=": "gte", "<": "lt", "<=": "lte"}
_rev_cmp_operators = {">": "<", ">=": "<=", "<": ">", "<=": "=>"}
_has_operators = {"ALL": "must", "ANY": "should"}
_length_quantities = {
    "elements": "nelements",
    "elements_rations": "nelements",
    "dimension_types": "dimension_types",
}


class MQuantity(Quantity):
    """A specialized version of
    optimade.filtertransformers.elasticsearch.Quantity for handling material
    queries. Added support for nested properties and converter functions.

    Attributes:
        name: The name of the quantity as used in the filter expressions.
        es_field: The name of the field for this quanity in elastic search, will be
            ``name`` by default.
        elastic_mapping_type: A decendent of an elasticsearch_dsl Field that denotes which
            mapping was used in the elastic search index.
        length_quantity: Elasticsearch does not support length of arrays, but we can
            map fields with array to other fields with ints about the array length. The
            LENGTH operator will only be supported for quantities with this attribute.
        has_only_quantity: Elasticsearch does not support exclusive search on arrays, like
            a list of chemical elements. But, we can order all elements by atomic number
            and use a keyword field with all elements to perform this search. This only
            works for elements (i.e. labels in ``CHEMICAL_SYMBOLS``) and quantities
            with this attribute.
        nested_quantity: To support optimade's 'zipped tuple' feature (e.g.
            'elements:elements_ratios HAS "H":>0.33), we use elasticsearch nested objects
            and nested queries. This quantity will provide the field for the nested
            object that contains the quantity (and others). The zipped tuples will only
            work for quantities that share the same nested object quantity.
        converter: Optional function that is used for manipulating searched
            values. Takes as input the token and returns the value that will be fed
            into the actual search.
        nested_path: The path used for nested queries. Provide only for nested
            fields.
    """

    def __init__(
        self,
        name,
        es_field: str = None,
        elastic_mapping_type: Field = None,
        length_quantity: "MQuantity" = None,
        has_only_quantity: "MQuantity" = None,
        nested_quantity: "MQuantity" = None,
        converter: Callable = None,
        nested_path: str = None,
    ):
        super().__init__(name, es_field, elastic_mapping_type, length_quantity, has_only_quantity, nested_quantity)
        self.converter = converter
        self.nested_path = nested_path


class MElasticTransformer(ElasticTransformer):
    '''
    A specialized Optimade/Lark transformer for handling material queries.
    Provides mostly the same functionality as
    optimade.filtertransformers.elasticsearch.ElasticTransformer, but has
    additions that make nested queries, parameter conversions and the use of
    boolean values possible.

    Uses elasticsearch_dsl and will produce a :class:`Q` instance.

    Arguments:
        quantities: A list of :class:`MQuantity`s that describe how optimade (and other)
            quantities are mapped to the elasticsearch index.
    '''
    def _query_op(self, quantity, op, value, nested=None):
        """
        Return a range, match, or term query for the given quantity, comparison
        operator, and value
        """
        field = self._field(quantity, nested=nested)
        if op in _cmp_operators:
            return Q("range", **{field: {_cmp_operators[op]: value}})

        if quantity.elastic_mapping_type == Text:
            query_type = "match"
        elif quantity.elastic_mapping_type in [Keyword, Integer, Boolean]:
            query_type = "term"
        else:
            raise NotImplementedError("Quantity has unsupported ES field type")

        if op in ["=", ""]:
            return Q(query_type, **{field: value})

        if op == "!=":
            return ~Q(  # pylint: disable=invalid-unary-operand-type
                query_type, **{field: value}
            )

    def _has_query_op(self, quantities, op, predicate_zip_list):
        """
        Returns a bool query that combines the operator queries ():func:`_query_op`)
        for each predicate and zipped quantity pericates combinations.
        """
        if op == "HAS":
            kind = "must"  # in case of HAS we do a must over the "list" of the one given element
        elif op == "HAS ALL":
            kind = "must"
        elif op == "HAS ANY":
            kind = "should"
        elif op == "HAS ONLY":
            # HAS ONLY comes with heavy limitations, because there is no such thing
            # in elastic search. Only supported for elements, where we can construct
            # an anonymous "formula" based on elements sorted by order number and
            # can do a = comparision to check if all elements are contained
            if len(quantities) > 1:
                raise Exception("HAS ONLY is not supported with zip")
            quantity = quantities[0]

            if quantity.has_only_quantity is None:
                raise Exception("HAS ONLY is not supported by %s" % quantity.name)

            def values():
                for predicates in predicate_zip_list:
                    if len(predicates) != 1:
                        raise Exception("Tuples not supported in HAS ONLY")
                    op, value = predicates[0]
                    if op != "=":
                        raise Exception("Predicated not supported in HAS ONLY")
                    if not isinstance(value, str):
                        raise Exception("Only strings supported in HAS ONLY")
                    yield value

            try:
                # Instead of the Optimade standard, the elements are combined
                # by the standard used by NOMAD.
                species, counts = get_hill_decomposition(list(values()))
                value = " ".join(["{}{}".format(s, "" if c == 1 else c) for s, c in zip(species, counts)])
            except KeyError:
                raise Exception("HAS ONLY is only supported for chemical symbols")

            return Q("term", **{quantity.has_only_quantity.name: value})
        else:
            raise NotImplementedError

        queries = [
            self._has_query(quantities, predicates) for predicates in predicate_zip_list
        ]
        args = {kind: queries}

        # Explicitly request that at least one term must match for the HAS ANY
        # query. ElasticSearch may either default to 0 or 1 depending on the
        # other search parameters:
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-bool-query.html#bool-min-should-match
        if kind == "should":
            args["minimum_should_match"] = 1
        return Q("bool", **args)

    @v_args(inline=True)
    def property_first_comparison(self, quantity, query):
        # property_first_comparison: property *_rhs

        # If the quantity corresponds to a nested ES value, the query is
        # wrapped with the appropriate outer nested query.
        inner_query = query(quantity)
        if quantity.nested_path:
            return Q("nested", path=quantity.nested_path, query=inner_query)
        return inner_query

    @v_args(inline=True)
    def constant_first_comparison(self, value, op, quantity):
        # constant_first_comparison: constant OPERATOR ( non_string_value | ...not_implemented_string )
        if not isinstance(quantity, Quantity):
            raise Exception("Only quantities can be compared to constant values.")

        # If the quantity corresponds to a nested ES value, the query is
        # wrapped with the appropriate outer nested query.
        inner_query = self._query_op(quantity, _rev_cmp_operators[op], value)
        if quantity.nested_path:
            return Q("nested", path=quantity.nested_path, query=inner_query)
        return inner_query

    @v_args(inline=True)
    def value_op_rhs(self, op, value):
        # value_op_rhs: OPERATOR value
        return lambda quantity: self._query_op(
            quantity,
            op,
            value if quantity.converter is None else quantity.converter(value)
        )

    @v_args(inline=True)
    def boolean(self, value):
        if value == "TRUE":
            return True
        if value == "FALSE":
            return False



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import re
import logging
from scipy.io.netcdf import netcdf_file
from ase.data import atomic_names
from ase import lattice as aselattice, Atoms

from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser import FileParser, TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, BasisSetAtomCentered, Method, XCFunctionals,\
    System, SingleConfigurationCalculation


class NCParser(FileParser):
    def __init__(self):
        super().__init__()
        self._configuration_types = ['MoleculeConfiguration', 'BulkConfiguration']
        self._units = dict(angstrom=ureg.angstrom, bohr=ureg.bohr, kelvin=ureg.kelvin)
        self._fingerprints = dict()

    @property
    def netcdf(self):
        if self._file_handler is None:
            try:
                self._file_handler = netcdf_file(self.mainfile, mmap=True)
            except Exception:
                return
            # prepare fingerprints required for variables
            if hasattr(self._file_handler, 'fingerprint_table'):
                fprints = [p.split(':') for p in self._file_handler.fingerprint_table.decode().split('#') if p]  # pylint: disable=maybe-no-member
                self._fingerprints = {p[1]: p[0] for p in fprints}

        return self._file_handler

    def resolve_unit(self, val):
        val = val.split('*')
        if len(val) == 2:
            return (float(val[0]) * self._units.get(val[1].lower(), ureg.angstrom)).to('angstrom').magnitude
        return float(val[0])

    def _resolve_configuration(self, data):

        # TODO implement UnitCell, ghost atoms

        lattices = dict(
            FaceCenteredCubic=aselattice.FCC, BodyCenteredCubic=aselattice.BCC, Triclinic=aselattice.TRI)

        re_f = r'[\d\.\-\+Ee]+'

        data = data.data[:].copy().tobytes().decode()

        elements = re.search(r'elements = \[(.+)\]', data)
        if not elements:
            return

        coordinates = re.search(r'coordinates *\= *(\[\s*\[[\s\S]+?\]\s*\])', data)
        if not coordinates:
            return

        try:
            numbers = [atomic_names.index(e.strip().title()) for e in elements.group(1).split(',')]

            positions = np.array([v.split(',') for v in re.findall(
                rf'\[( *{re_f} *\, *{re_f} *\, *{re_f} *)\]', coordinates.group(1))], dtype=np.dtype(np.float64))

            atoms = Atoms(numbers=numbers, positions=positions)

        except Exception:
            return

        velocities = re.search(r'velocities *\= *(\[\s*\[[\s\S]+?\]\s*\])', data)
        if velocities:
            atoms.set_velocities(np.array([v.split(',') for v in re.findall(
                rf'\[( *{re_f} *\, *{re_f} *\, *{re_f} *)\]', velocities.group(1))], dtype=np.dtype(np.float64)))

        if 'MoleculeConfiguration' in data:
            return atoms

        atoms.set_pbc(True)

        lattice = re.search(r'\nlattice = (\w+) *\((.+)\)', data)
        lattice, parameters = lattice.groups() if lattice else ('', '')
        parameters = [self.resolve_unit(p) for p in re.findall(rf'({re_f} *\* *\w+)', parameters)]
        lattice = lattices.get(lattice)
        if lattice is None:
            return

        try:
            atoms.set_cell(lattice(*parameters).tocell(), scale_atoms='fractional' in data)
        except Exception:
            pass

        return atoms

    def parse(self, key):
        val = None
        if hasattr(self.netcdf, key):
            val = getattr(self.netcdf, key)
            if isinstance(val, bytes):
                val = val.decode()
        elif key == 'configuration_names':
            val = []
            re_name = re.compile(r'(\w+Configuration\_gID\d+)$')
            for name in self.netcdf.variables.keys():
                name = re.match(re_name, name)
                if name:
                    val.append(name.group(1))
        elif key == 'atoms':
            val = dict()
            for name in self.get('configuration_names', []):
                data = self.netcdf.variables.get(name)
                if data is None:
                    continue
                val[name] = self._resolve_configuration(data)
        elif key == 'parameters':
            val = dict()
            for name in self.get('configuration_names', []):
                data = self.netcdf.variables.get('%s_calculator' % name)
                if data is None:
                    continue
                val[name] = data.data.tobytes()
        elif key == 'energies':
            val = dict()
            # energy_keys = [k for k in self.netcdf.variables.keys() if k.startswith('TotalEnergy')]
            energy_re = r'TotalEnergy\_(gID\d+)\_component\_(\S+)'
            for key_n in self.netcdf.variables.keys():
                energy_key = re.match(energy_re, key_n)
                if energy_key:
                    fp = self.netcdf.variables[
                        'TotalEnergy_%s_finger_print' % energy_key.group(1)].data.tobytes().decode()
                    val.setdefault(fp, {})
                    val[fp][energy_key.group(2)] = self.netcdf.variables[key_n].data[0] * ureg.eV
        elif key == 'forces':
            val = dict()
            forces_re = r'Forces\_(gID\d+)\_atom\_resolved\_forces'
            for key_n in self.netcdf.variables.keys():
                forces_key = re.match(forces_re, key_n)
                if forces_key:
                    fp = self.netcdf.variables[
                        'TotalEnergy_%s_finger_print' % forces_key.group(1)].data.tobytes().decode()
                    val[fp] = self.netcdf.variables[key_n].data * (ureg.eV / ureg.angstrom)

        # TODO implement stress, bandstructure, eigenvalues
        self._results[key] = val


class CalculatorParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity('smearing_width', r'electron_temperature *\= *([\d\.]+)', dtype=float),
            Quantity('charge', r'charge *\= *([\d\.]+)', dtype=float),
            Quantity('xc_functional', r'exchange_correlation *\= *(\S+)')
        ]


class ATKParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/atk', code_name='AtomistixToolKit',
            code_homepage='https://www.synopsys.com/silicon/quantumatk.html',
            mainfile_name_re=r'^.*\.nc', mainfile_mime_re=r'application/octet-stream')
        self.nc_parser = NCParser()
        self.calculator_parser = CalculatorParser()

        self._metainfo_map = {
            'Exchange-Correlation': 'energy_XC', 'Kinetic': 'electronic_kinetic_energy',
            'Entropy-Term': 'energy_correction_entropy', 'Electrostatic': 'energy_electrostatic'}

        self._xc_functional_map = {
            'LDA.RPA': ['LDA_X', 'LDA_C_RPA'],
            'LDA.PZ': ['LDA_X', 'LDA_C_PZ'],
            'LDA.PW': ['LDA_X', 'LDA_C_PW'],
            'GGA.PW91': ['GGA_X_PW91', 'GGA_C_PW91'],
            'GGA.PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'GGA.PBES': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            'GGA.RPBE': ['GGA_X_RPBE', 'GGA_C_PBE'],
            'BLYP': ['GGA_X_B88', 'GGA_C_LYP'],
            'HCTH407': ['GGA_XC_HCTH_407'],
            'WC': ['GGA_X_WC', 'GGA_C_PBE'],
            'AM05': ['GGA_X_AM05', 'GGA_C_AM05'],
            'mBEEF': ['MGGA_X_MBEEF', 'GGA_C_PBE_SOL']}

    def init_parser(self):
        self.nc_parser.mainfile = self.filepath
        self.nc_parser.logger = self.logger

    def parse_configurations(self):
        sec_run = self.archive.section_run[0]

        def parse_method(name):
            sec_method = sec_run.m_create(Method)

            sec_method.relativity_method = 'pseudo_scalar_relativistic'
            sec_method.electronic_structure_method = 'DFT'

            sec_method.smearing_kind = 'fermi'

            parameters = self.nc_parser.get('parameters').get(name)
            if parameters is None:
                return sec_method

            # this is only a dummy filename as we will get the info from the calculator data
            self.calculator_parser.mainfile = self.filepath
            self.calculator_parser._file_handler = parameters

            for key, val in self.calculator_parser.items():
                if val is not None:
                    setattr(sec_method, key, val)

            for xc_functional in self._xc_functional_map.get(self.calculator_parser.get('xc_functional'), []):
                sec_xc = sec_method.m_create(XCFunctionals)
                sec_xc.XC_functional_name = xc_functional

            return sec_method

        def parse_system(atoms):
            if atoms is None:
                return

            sec_system = sec_run.m_create(System)
            sec_system.atom_labels = atoms.get_chemical_symbols()
            sec_system.atom_positions = atoms.get_positions() * ureg.angstrom
            sec_system.lattice_vectors = atoms.get_cell().array * ureg.angstrom
            sec_system.configuration_periodic_dimensions = atoms.get_pbc()
            velocities = atoms.get_velocities()
            if velocities:
                sec_system.atom_velocities = velocities * (ureg.angstrom / ureg.fs)

            return sec_system

        def parse_scc(fingerprint):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)

            # energies
            energy_total = 0.
            for key, val in self.nc_parser.get('energies').get(fingerprint, {}).items():
                key = self._metainfo_map.get(key)
                energy_total += val
                if key is not None:
                    setattr(sec_scc, key, val)
            sec_scc.energy_total = energy_total

            # forces
            forces = self.nc_parser.get('forces').get(fingerprint)
            if forces is not None:
                sec_scc.atom_forces = forces

            return sec_scc

        for name, atoms in self.nc_parser.get('atoms', {}).items():
            sec_system = parse_system(atoms)
            sec_method = parse_method(name)
            fingerprint = self.nc_parser._fingerprints.get('gID%s' % name.split('gID')[-1])
            sec_scc = parse_scc(fingerprint)
            if sec_system is not None:
                sec_scc.single_configuration_calculation_to_system_ref = sec_system
            if sec_method is not None:
                sec_scc.single_configuration_to_calculation_method_ref = sec_method

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'ATK'
        version = self.nc_parser.get('version')
        sec_run.program_version = version if version is not None else 'unavailable'
        sec_run.program_basis_set_type = 'numeric AOs'

        sec_basis = sec_run.m_create(BasisSetAtomCentered)
        sec_basis.basis_set_atom_centered_short_name = 'ATK LCAO basis'

        self.parse_configurations()
